

def funPrettyPlotDual(x,y, dy_dx):
    yLower1 = min(y)
    yLower2 = min(dy_dx)
    
    yUpper1 = max(y)
    yUpper2 = max(dy_dx)
    
    #Establish upper and lower bounds for y-axis
    yLower = min(yLower1, yLower2)
    yUpper = max(yUpper1, yUpper2)
    
    xLower = min(x)
    xUpper = max(x)
    
    if yUpper - yLower > 100:
        d = yUpper - yLower
        yLower = (yLower/d)*25
        yUpper = (yUpper/d)*25

    from matplotlib import pyplot as plt
    # Define plots.
    plt.figure(figsize=(6, 4)) # Scale
    plt.plot(x, y, label=r'$\operatorname{Re}[f(x + \epsilon)]$')
    plt.plot(x, dy_dx, label=r'$\operatorname{Im}[f(x + \epsilon)]$')

# Configure axis properties.
    axes = plt.gca()
    axes.set_xlim([xLower, xUpper])
    axes.set_ylim([yLower, yUpper])

# Configure plot properties.
    plt.xlabel(r'x')
    plt.grid(True)
    plt.legend(loc='upper right')
    plt.tight_layout()
    plt.show()
    
    return

def funPrettyPlot(x,y, dy_dx):
    yLower1 = min(y)
    yLower2 = min(dy_dx)
    
    yUpper1 = max(y)
    yUpper2 = max(dy_dx)
    
    #Establish upper and lower bounds for y-axis
    yLower = min(yLower1, yLower2)
    yUpper = max(yUpper1, yUpper2)
    
    xLower = min(x)
    xUpper = max(x)
    
    
    if yUpper - yLower > 100:
       d = yUpper - yLower
       yLower = (yLower/d)*25
       yUpper = (yUpper/d)*25



    from matplotlib import pyplot as plt
    # Define plots.
    plt.figure(figsize=(6, 4)) # Scale
    plt.plot(x, y, label=r'$\operatorname{Re}[f(x + ih)]$')
    plt.plot(x, dy_dx, label=r'$\frac{\operatorname{Im}[f(x + ih)]}{h}$')

# Configure axis properties.
    axes = plt.gca()
    axes.set_xlim([xLower, xUpper])
    axes.set_ylim([yLower, yUpper])

# Configure plot properties.
    plt.xlabel(r'x')
    plt.grid(True)
    plt.legend(loc='upper right')
    plt.tight_layout()
    plt.show()
    
    return

    
def funPlotStepSize(y,h, x0, exactDx, machinePrecision):
    
    import numpy as np
    from matplotlib import pyplot as plt
    
#Implements CTSE method
    def CTSE(fx,x0,h):
     x0 = x0 + h*1j
     y = fx(x0)
     f_x0 = y.real
     dy_dx = y.imag/h
     return dy_dx 
    
    def fd(fx,x0,h):
        dy_dx = (fx(x0 + h) - fx(x0))/h
        return dy_dx
    
    def cd(fx,x0,h):
        dy_dx = (fx(x0 + h) - fx(x0 - h))/(2*h)
        return dy_dx
    

    forwardDifference = []
    centralDifference = []
    ctse = []
    # Compute f'(x) using forward and central difference, and CTSE.
    for i in h:
        nStepSize = 10**-i
        forwardDifference.append(fd(y,x0,nStepSize ))
        centralDifference.append(cd(y,x0,nStepSize ))
        ctse.append(CTSE(y,x0,nStepSize))
    
    # Compute relative errors of the above methods.
    relativeErrorForwardDiff = abs((forwardDifference - exactDx) / exactDx)
    relativeErrorCentralDiff = abs((centralDifference - exactDx) / exactDx)
    relativeErrorCtse = abs((ctse - exactDx) / exactDx)
    
    # Check if a value in list "relativeErrorCtse" is more precise than
    # machine precision, if so, truncate value to machine precision.
    for i, x in enumerate(relativeErrorCtse):
        if x < machinePrecision:
            relativeErrorCtse[i] = machinePrecision
    
    # Plot and configure figures.
    plt.figure(figsize=(7, 4))
    plt.plot(
        h,
        relativeErrorForwardDiff,
        label='Forward Difference',
        marker="^",
        color="blue",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCentralDiff,
        label='Central Difference',
        marker="s",
        color="purple",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCtse,
        label='CTSE',
        marker="o",
        color="red",
        markerfacecolor="none")
    axes = plt.gca()
    axes.set_xlim([0, 30])
    axes.set_ylim([10**-17, 10**1])
    plt.xlabel('Step size h')
    plt.ylabel('Relative Error')
    xticksOriginal = [0, 5, 10, 15, 20, 25, 30]
    xticksNew = [
        '$10^{0}$',
        '$10^{-5}$',
        '$10^{-10}$',
        '$10^{-15}$',
        '$10^{-20}$',
        '$10^{-25}$',
        '$10^{-30}$']
    plt.xticks(xticksOriginal, xticksNew)
    plt.grid(True)
    plt.legend()
    plt.tight_layout()
    plt.yscale("log")
    plt.show()
    return


def funPlotStepSizeDual(y,y_dual, h, x0, exactDx, machinePrecision):
    
    import numpy as np
    from matplotlib import pyplot as plt
    import multiZ.symdual as sd
    import sympy as sp
    import multiZ.symdual as sd
    
#Implements CTSE method
    def CTSE(fx,x0,h):
     x0 = x0 + h*1j
     y = fx(x0)
     f_x0 = y.real
     dy_dx = y.imag/h
     return dy_dx 
    
    def fd(fx,x0,h):
        dy_dx = (fx(x0 + h) - fx(x0))/h
        return dy_dx
    
    def cd(fx,x0,h):
        dy_dx = (fx(x0 + h) - fx(x0 - h))/(2*h)
        return dy_dx
    
    def Dual(y, x0, ss):
        
        h = sp.symbols('h')
        x = sp.symbols('x')
        fx = y.real
        dy_dx = (y.imag)/h
        dy_dx = sp.lambdify([x,h], dy_dx, "numpy") 
        dy_dx = dy_dx(x0,ss)
        


        return dy_dx
    

    forwardDifference = []
    centralDifference = []
    ctse = []
    dual = []
    # Compute f'(x) using forward and central difference, and CTSE.
    for i in h:
        nStepSize = 10**-i
        forwardDifference.append(fd(y,x0,nStepSize ))
        centralDifference.append(cd(y,x0,nStepSize ))
        ctse.append(CTSE(y,x0,nStepSize))
        dual.append(Dual(y_dual, x0,nStepSize))
        
    
    # Compute relative errors of the above methods.
    relativeErrorForwardDiff = abs((forwardDifference - exactDx) / exactDx)
    relativeErrorCentralDiff = abs((centralDifference - exactDx) / exactDx)
    relativeErrorCtse = abs((ctse - exactDx) / exactDx)
    relativeErrorDual = abs((dual - exactDx) / exactDx)
    
    # Check if a value in list "relativeErrorCtse" is more precise than
    # machine precision, if so, truncate value to machine precision.
    for i, x in enumerate(relativeErrorCtse):
        if x < machinePrecision:
            relativeErrorCtse[i] = machinePrecision
    for i, x in enumerate(relativeErrorDual):
            if x < machinePrecision:
                relativeErrorDual[i] = machinePrecision
    
    # Plot and configure figures.
    plt.figure(figsize=(7, 4))
    plt.plot(
        h,
        relativeErrorForwardDiff,
        label='Forward Difference',
        marker="^",
        color="blue",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCentralDiff,
        label='Central Difference',
        marker="s",
        color="purple",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCtse,
        label='CTSE',
        marker="o",
        color="red",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorDual,
        label='Dual',
        marker="o",
        color="green",
        markerfacecolor="none")
    axes = plt.gca()
    axes.set_xlim([0, 30])
    axes.set_ylim([10**-17, 10**1])
    plt.xlabel('Step size h')
    plt.ylabel('Relative Error')
    xticksOriginal = [0, 5, 10, 15, 20, 25, 30]
    xticksNew = [
        '$10^{0}$',
        '$10^{-5}$',
        '$10^{-10}$',
        '$10^{-15}$',
        '$10^{-20}$',
        '$10^{-25}$',
        '$10^{-30}$']
    plt.xticks(xticksOriginal, xticksNew)
    plt.grid(True)
    plt.legend()
    plt.tight_layout()
    plt.yscale("log")
    plt.show()
    return

def funPlotStepSize_NI(y,h, a_int, b_int, c_param, sInterestedParam, exactDa, exactDb, exactDc, machinePrecision):
    
    import numpy as np
    from matplotlib import pyplot as plt
    import sympy as sym 
    
    #Define the approximation to the integral 
    def I_abc(f,a,b,c):
      y = ((b-a)/8)*(f(a,c) + 3*f((2*a + b)/3,c) + 3*f((a + 2*b)/3,c)  + f(b,c))
      return y
    
    
    
    #Implements CTSE method
    def CTSE(f, a,b,c,h):
     a_pert = a + h*1j
     b_pert = b + h*1j
     c_pert = c + h*1j
     y1 = I_abc(f, a_pert, b, c)
     y2 = I_abc(f, a, b_pert, c)
     y3 = I_abc(f, a,b,c_pert)
     y = y1.real
     dy_da = y1.imag/h
     dy_db = y2.imag/h
     dy_dc = y3.imag/h
     return dy_da, dy_db, dy_dc, y
    
    #Implements FD method
    def fd(f, a,b,c,h):
        dy_da = (I_abc(f, a + h,b,c) - I_abc(f, a,b,c))/h
        dy_db = (I_abc(f, a,b + h,c) - I_abc(f, a,b,c))/h
        dy_dc = (I_abc(f, a,b,c + h) - I_abc(f, a,b,c))/h
        return dy_da, dy_db, dy_dc
    
    #Implements CD method
    def cd(f, a,b,c,h):
        dy_da = (I_abc(f, a + h,b,c) - I_abc(f, a - h,b,c))/(2*h)
        dy_db = (I_abc(f, a,b + h,c) - I_abc(f, a,b - h,c))/(2*h)
        dy_dc = (I_abc(f, a,b,c + h) - I_abc(f, a,b,c - h))/(2*h)
        return dy_da, dy_db, dy_dc
    

    forwardDifference = []
    centralDifference = []
    ctse = []
    
    # Compute f'(x) using forward and central difference, and CTSE.
    for i in h:
        nStepSize = 10**-i
        forwardDifference.append(fd(y, a_int,b_int,c_param,nStepSize))
        centralDifference.append(cd(y, a_int,b_int,c_param,nStepSize ))
        ctse.append(CTSE(y, a_int,b_int,c_param,nStepSize))
        
        
    if sInterestedParam == 'a':
        nChoice = 0
        nExact = exactDa
    if sInterestedParam == 'b':
        nChoice = 1
        nExact = exactDb
    if sInterestedParam == 'c':
        nChoice = 2
        nExact = exactDc
    ctse = np.array(ctse)
    forwardDifference = np.array(forwardDifference)
    centralDifference = np.array(centralDifference)
    ctse = ctse[:,nChoice]
    forwardDifference= forwardDifference[:,nChoice]
    centralDifference= centralDifference[:,nChoice]
 
    # Compute relative errors of the above methods.
    relativeErrorForwardDiff = abs((forwardDifference - nExact) / nExact)
    relativeErrorCentralDiff = abs((centralDifference- nExact) / nExact)
    relativeErrorCtse = abs((ctse  - nExact) / nExact)
    
    mMinVals = [min(relativeErrorForwardDiff ), min(relativeErrorCentralDiff), min(relativeErrorCtse)]
    mMaxVals = [max(relativeErrorForwardDiff ), max(relativeErrorCentralDiff), max(relativeErrorCtse)]
    nMinyVals = min(mMinVals)*10**(-1)
    nMaxyVals = max(mMaxVals)
    # Check if a value in list "relativeErrorCtse" is more precise than
    # machine precision, if so, truncate value to machine precision.
    for i, x in enumerate(relativeErrorCtse):
        if x < machinePrecision:
            relativeErrorCtse[i] = machinePrecision
 
    # Plot and configure figures.
    plt.figure(figsize=(7, 4))
    plt.plot(
        h,
        relativeErrorForwardDiff,
        label='Forward Difference',
        marker="^",
        color="blue",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCentralDiff,
        label='Central Difference',
        marker="s",
        color="purple",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCtse,
        label='CTSE',
        marker="o",
        color="red",
        markerfacecolor="none")
    axes = plt.gca()
    axes.set_xlim([0, 30])
    axes.set_ylim([nMinyVals, nMaxyVals])
    plt.xlabel('Step size h')
    plt.ylabel('Relative Error')
    xticksOriginal = [0, 5, 10, 15, 20, 25, 30]
    xticksNew = [
        '$10^{0}$',
        '$10^{-5}$',
        '$10^{-10}$',
        '$10^{-15}$',
        '$10^{-20}$',
        '$10^{-25}$',
        '$10^{-30}$']
    plt.xticks(xticksOriginal, xticksNew)
    plt.grid(True)
    plt.legend()
    plt.tight_layout()
    plt.yscale("log")
    plt.show()
    
def funPlotStepSize_NI_GL(y,h, a_int, b_int, c_param, sInterestedParam, exactDa, exactDb, exactDc, machinePrecision, w, xi):
    
    import numpy as np
    from matplotlib import pyplot as plt
    import sympy as sym 
    
    #Define the approximation to the integral 
    def I_abc(f,a,b,c, w, xi):
      temp = 0
      for i in range(0,len(w)):
        temp = temp + w[i]*f(((b-a)/2)*xi[i] + ((b+a)/2), c)
      y = ((b - a)/2)*(temp)
      return y
    
    
    
     #Implements CTSE method
    def CTSE(f, a,b,c,h,w, xi):
     a_pert = a + h*1j
     b_pert = b + h*1j
     c_pert = c + h*1j
     y1 = I_abc(f, a_pert, b, c,w,xi)
     y2 = I_abc(f, a, b_pert, c,w,xi)
     y3 = I_abc(f, a,b,c_pert, w, xi)
     y = y1.real
     dy_da = y1.imag/h
     dy_db = y2.imag/h
     dy_dc = y3.imag/h
     return dy_da, dy_db, dy_dc, y
    
    def fd(f, a,b,c,h,w, xi):
        dy_da = (I_abc(f, a + h,b,c, w, xi) - I_abc(f, a,b,c, w, xi))/h
        dy_db = (I_abc(f, a,b + h,c, w, xi) - I_abc(f, a,b,c, w, xi))/h
        dy_dc = (I_abc(f, a,b,c + h, w, xi) - I_abc(f, a,b,c, w, xi))/h
        return dy_da, dy_db, dy_dc
    
    def cd(f, a,b,c,h, w, xi):
        dy_da = (I_abc(f, a + h,b,c,w ,xi) - I_abc(f, a - h,b,c, w, xi))/(2*h)
        dy_db = (I_abc(f, a,b + h,c, w, xi) - I_abc(f, a,b - h,c, w, xi))/(2*h)
        dy_dc = (I_abc(f, a,b,c + h,w, xi) - I_abc(f, a,b,c - h, w, xi))/(2*h)
        return dy_da, dy_db, dy_dc
    

    forwardDifference = []
    centralDifference = []
    ctse = []
    
    # Compute f'(x) using forward and central difference, and CTSE.
    for i in h:
        nStepSize = 10**-i
        forwardDifference.append(fd(y, a_int,b_int,c_param,nStepSize, w, xi))
        centralDifference.append(cd(y, a_int,b_int,c_param,nStepSize, w, xi ))
        ctse.append(CTSE(y, a_int,b_int,c_param,nStepSize, w, xi))
        
        
    if sInterestedParam == 'a':
        nChoice = 0
        nExact = exactDa
    if sInterestedParam == 'b':
        nChoice = 1
        nExact = exactDb
    if sInterestedParam == 'c':
        nChoice = 2
        nExact = exactDc
    ctse = np.array(ctse)
    forwardDifference = np.array(forwardDifference)
    centralDifference = np.array(centralDifference)
    ctse = ctse[:,nChoice]
    forwardDifference= forwardDifference[:,nChoice]
    centralDifference= centralDifference[:,nChoice]
 
    # Compute relative errors of the above methods.
    relativeErrorForwardDiff = abs((forwardDifference - nExact) / nExact)
    relativeErrorCentralDiff = abs((centralDifference- nExact) / nExact)
    relativeErrorCtse = abs((ctse  - nExact) / nExact)
    
    mMinVals = [min(relativeErrorForwardDiff ), min(relativeErrorCentralDiff), min(relativeErrorCtse)]
    mMaxVals = [max(relativeErrorForwardDiff ), max(relativeErrorCentralDiff), max(relativeErrorCtse)]
    nMinyVals = min(mMinVals)*10**(-1)
    nMaxyVals = max(mMaxVals)
    # Check if a value in list "relativeErrorCtse" is more precise than
    # machine precision, if so, truncate value to machine precision.
    for i, x in enumerate(relativeErrorCtse):
        if x < machinePrecision:
            relativeErrorCtse[i] = machinePrecision
 
    # Plot and configure figures.
    plt.figure(figsize=(7, 4))
    plt.plot(
        h,
        relativeErrorForwardDiff,
        label='Forward Difference',
        marker="^",
        color="blue",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCentralDiff,
        label='Central Difference',
        marker="s",
        color="purple",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCtse,
        label='CTSE',
        marker="o",
        color="red",
        markerfacecolor="none")
    axes = plt.gca()
    axes.set_xlim([0, 30])
    axes.set_ylim([nMinyVals, nMaxyVals])
    plt.xlabel('Step size h')
    plt.ylabel('Relative Error')
    xticksOriginal = [0, 5, 10, 15, 20, 25, 30]
    xticksNew = [
        '$10^{0}$',
        '$10^{-5}$',
        '$10^{-10}$',
        '$10^{-15}$',
        '$10^{-20}$',
        '$10^{-25}$',
        '$10^{-30}$']
    plt.xticks(xticksOriginal, xticksNew)
    plt.grid(True)
    plt.legend()
    plt.tight_layout()
    plt.yscale("log")
    plt.show()    
def funCalcSymbolicFunctions_NI(y,x,a,b,c, a_int, b_int, c_param):
    import sympy as sym
    I = sym.integrate(y,(x,a,b))
    dI_da = sym.diff(I,a)
    dI_db = sym.diff(I,b)
    dI_dc = sym.diff(I,c)
    dI_da = sym.lambdify([a,c], dI_da, 'numpy')
    dI_db = sym.lambdify([b,c], dI_db, 'numpy')
    dI_dc = sym.lambdify([a,b,c], dI_dc, 'numpy')
    exactDa = dI_da(a_int,c_param)
    exactDb = dI_db(b_int,c_param)
    exactDc = dI_dc(a_int,b_int,c_param)
    I = sym.lambdify([a,b,c], I, 'numpy')
    exactFx = I(a_int,b_int,c_param)
    
    
    return  exactDa, exactDb, exactDc, exactFx

# Comment: Should we use given NR subroutines from numpy/scipy 
def NR_method(x0, nIter, nTol, h, f):
  import numpy as np
  from matplotlib import pyplot as plt
  import sympy as sym 
  mVals = []
  mError = []
  for i in range(0,nIter):
    if i == 0:
      x_i = x0 - h*((f(x0 + h*1j)).real)/(f(x0 + h*1j).imag)
      mVals.append(x_i)
      mError.append(np.abs(f(x_i)))
    else:
      x_i =  x_i - h*(((f(x_i + h*1j)).real)/(f(x_i + h*1j).imag))
      mVals.append(x_i)
      mError.append(np.abs(f(x_i)))
      if np.abs(f(x_i)) < nTol:
        break
  plt.figure(figsize=(7, 4))
  plt.plot(mError, marker="o", label = '$|f(x_j)|$',
        color="red",
        markerfacecolor="none")
  axes = plt.gca()
  if i < 10:
    ticks = [j for j in range(0,len(mError))]
    axes.set_xticks(ticks)
  plt.xlabel('Number of iterations')
  plt.ylabel('Forward Error')
  plt.yscale("log")
  plt.grid(True)
  plt.tight_layout()
  plt.hlines(nTol, 0, i, colors='b', linestyles='solid', label='Tolerence')
  plt.legend()
  return



# Comment: Should we use given NR subroutines from numpy/scipy?
def NewtonRaphson_Complex(zf, x0, max_iter=50, tol=1e-6, step_size=1e-10, full_output=False ):
    """
    Find a zeros of a real function using the Newton-Raphson (NR) with CTSE.

    INPUTS:
    - zf:          callable, Residual function.
    - x0:          float, Initial guess.
    - max_iter:    integer, optional, maximum number of iterations to be performed.
    - tol:         float, optional, tolerance for the residual to terminate the NR loop.
    - step_size:   float, optional, step-size to be used for the CTSE calculations.
    - full_output: bool, optional, request full output.

    RESULT:
    The result of this function is a tuple of values that include:
    - x: real, root of the algorithm.
    - x_iter: ndarray, optional if full_output==True, evolution of the root during the iterations.
    - f_iter: ndarray, evolution of the residual.

    EXAMPLE:

    >>> import numpy as np
    >>> import utils
    >>> 
    >>> def residual(x):
    >>>     return np.cos(x)-x
    >>> # end
    >>> 
    >>> utils.NewtonRaphson_Complex(residual,)

    """
    import numpy as np
    # from matplotlib import pyplot as plt
    # import sympy as sym 
    mVals = []
    mError = []
    for i in range(0,nIter):
        if i == 0:
            x_i = x0 - h*((f(x0 + h*1j)).real)/(f(x0 + h*1j).imag)
            mVals.append(x_i)
            mError.append(np.abs(f(x_i)))
        else:
            x_i =  x_i - h*(((f(x_i + h*1j)).real)/(f(x_i + h*1j).imag))
            mVals.append(x_i)
            mError.append(np.abs(f(x_i)))
            if np.abs(f(x_i)) < nTol:
                break
    plt.figure(figsize=(7, 4))
    plt.plot(mError, marker="o", label = '$|f(x_j)|$',
        color="red",
        markerfacecolor="none")
    axes = plt.gca()
    if i < 10:
    ticks = [j for j in range(0,len(mError))]
    axes.set_xticks(ticks)
    plt.xlabel('Number of iterations')
    plt.ylabel('Forward Error')
    plt.yscale("log")
    plt.grid(True)
    plt.tight_layout()
    plt.hlines(nTol, 0, i, colors='b', linestyles='solid', label='Tolerence')
    plt.legend()
    return