

def funPrettyPlotDual(x,y, dy_dx):
    yLower1 = min(y)
    yLower2 = min(dy_dx)
    
    yUpper1 = max(y)
    yUpper2 = max(dy_dx)
    
    #Establish upper and lower bounds for y-axis
    yLower = min(yLower1, yLower2)
    yUpper = max(yUpper1, yUpper2)
    
    xLower = min(x)
    xUpper = max(x)
    
    if yUpper - yLower > 100:
        d = yUpper - yLower
        yLower = (yLower/d)*25
        yUpper = (yUpper/d)*25

    from matplotlib import pyplot as plt
    # Define plots.
    plt.figure(figsize=(6, 4)) # Scale
    plt.plot(x, y, label=r'$\operatorname{Re}[f(x + \epsilon)]$')
    plt.plot(x, dy_dx, label=r'$\operatorname{Im}[f(x + \epsilon)]$')

# Configure axis properties.
    axes = plt.gca()
    axes.set_xlim([xLower, xUpper])
    axes.set_ylim([yLower, yUpper])

# Configure plot properties.
    plt.xlabel(r'x')
    plt.grid(True)
    plt.legend(loc='upper right')
    plt.tight_layout()
    plt.show()
    
    return

def funPrettyPlot(x,y, dy_dx):
    yLower1 = min(y)
    yLower2 = min(dy_dx)
    
    yUpper1 = max(y)
    yUpper2 = max(dy_dx)
    
    #Establish upper and lower bounds for y-axis
    yLower = min(yLower1, yLower2)
    yUpper = max(yUpper1, yUpper2)
    
    xLower = min(x)
    xUpper = max(x)
    
    
    if yUpper - yLower > 100:
       d = yUpper - yLower
       yLower = (yLower/d)*25
       yUpper = (yUpper/d)*25



    from matplotlib import pyplot as plt
    # Define plots.
    plt.figure(figsize=(6, 4)) # Scale
    plt.plot(x, y, label=r'$\operatorname{Re}[f(x + ih)]$')
    plt.plot(x, dy_dx, label=r'$\frac{\operatorname{Im}[f(x + ih)]}{h}$')

# Configure axis properties.
    axes = plt.gca()
    axes.set_xlim([xLower, xUpper])
    axes.set_ylim([yLower, yUpper])

# Configure plot properties.
    plt.xlabel(r'x')
    plt.grid(True)
    plt.legend(loc='upper right')
    plt.tight_layout()
    plt.show()
    
    return

    
def funPlotStepSize(y,h, x0, exactDx, machinePrecision):
    
    import numpy as np
    from matplotlib import pyplot as plt
    
#Implements CTSE method
    def CTSE(fx,x0,h):
     x0 = x0 + h*1j
     y = fx(x0)
     f_x0 = y.real
     dy_dx = y.imag/h
     return dy_dx 
    
    def fd(fx,x0,h):
        dy_dx = (fx(x0 + h) - fx(x0))/h
        return dy_dx
    
    def cd(fx,x0,h):
        dy_dx = (fx(x0 + h) - fx(x0 - h))/(2*h)
        return dy_dx
    

    forwardDifference = []
    centralDifference = []
    ctse = []
    # Compute f'(x) using forward and central difference, and CTSE.
    for i in h:
        nStepSize = 10**-i
        forwardDifference.append(fd(y,x0,nStepSize ))
        centralDifference.append(cd(y,x0,nStepSize ))
        ctse.append(CTSE(y,x0,nStepSize))
    
    # Compute relative errors of the above methods.
    relativeErrorForwardDiff = abs((forwardDifference - exactDx) / exactDx)
    relativeErrorCentralDiff = abs((centralDifference - exactDx) / exactDx)
    relativeErrorCtse = abs((ctse - exactDx) / exactDx)
    
    # Check if a value in list "relativeErrorCtse" is more precise than
    # machine precision, if so, truncate value to machine precision.
    for i, x in enumerate(relativeErrorCtse):
        if xe < machinePrecision:
            rlativeErrorCtse[i] = machinePrecision
    
    # Plot and configure figures.
    plt.figure(figsize=(7, 4))
    plt.plot(
        h,
        relativeErrorForwardDiff,
        label='Forward Difference',
        marker="^",
        color="blue",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCentralDiff,
        label='Central Difference',
        marker="s",
        color="purple",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCtse,
        label='CTSE',
        marker="o",
        color="red",
        markerfacecolor="none")
    axes = plt.gca()
    axes.set_xlim([0, 30])
    axes.set_ylim([10**-17, 10**1])
    plt.xlabel('Step size h')
    plt.ylabel('Relative Error')
    xticksOriginal = [0, 5, 10, 15, 20, 25, 30]
    xticksNew = [
        '$10^{0}$',
        '$10^{-5}$',
        '$10^{-10}$',
        '$10^{-15}$',
        '$10^{-20}$',
        '$10^{-25}$',
        '$10^{-30}$']
    plt.xticks(xticksOriginal, xticksNew)
    plt.grid(True)
    plt.legend()
    plt.tight_layout()
    plt.yscale("log")
    plt.show()
    return


def funPlotStepSizeDual(y,y_dual, h, x0, exactDx, machinePrecision):
    
    import numpy as np
    from matplotlib import pyplot as plt
    import multiZ.symdual as sd
    import sympy as sp
    import multiZ.symdual as sd
    
#Implements CTSE method
    def CTSE(fx,x0,h):
     x0 = x0 + h*1j
     y = fx(x0)
     f_x0 = y.real
     dy_dx = y.imag/h
     return dy_dx 
    
    def fd(fx,x0,h):
        dy_dx = (fx(x0 + h) - fx(x0))/h
        return dy_dx
    
    def cd(fx,x0,h):
        dy_dx = (fx(x0 + h) - fx(x0 - h))/(2*h)
        return dy_dx
    
    def Dual(y, x0, ss):
        
        h = sp.symbols('h')
        x = sp.symbols('x')
        fx = y.real
        dy_dx = (y.imag)/h
        dy_dx = sp.lambdify([x,h], dy_dx, "numpy") 
        dy_dx = dy_dx(x0,ss)
        


        return dy_dx
    

    forwardDifference = []
    centralDifference = []
    ctse = []
    dual = []
    # Compute f'(x) using forward and central difference, and CTSE.
    for i in h:
        nStepSize = 10**-i
        forwardDifference.append(fd(y,x0,nStepSize ))
        centralDifference.append(cd(y,x0,nStepSize ))
        ctse.append(CTSE(y,x0,nStepSize))
        dual.append(Dual(y_dual, x0,nStepSize))
        
    
    # Compute relative errors of the above methods.
    relativeErrorForwardDiff = abs((forwardDifference - exactDx) / exactDx)
    relativeErrorCentralDiff = abs((centralDifference - exactDx) / exactDx)
    relativeErrorCtse = abs((ctse - exactDx) / exactDx)
    relativeErrorDual = abs((dual - exactDx) / exactDx)
    
    # Check if a value in list "relativeErrorCtse" is more precise than
    # machine precision, if so, truncate value to machine precision.
    for i, x in enumerate(relativeErrorCtse):
        if x < machinePrecision:
            relativeErrorCtse[i] = machinePrecision
    for i, x in enumerate(relativeErrorDual):
            if x < machinePrecision:
                relativeErrorDual[i] = machinePrecision
    
    # Plot and configure figures.
    plt.figure(figsize=(7, 4))
    plt.plot(
        h,
        relativeErrorForwardDiff,
        label='Forward Difference',
        marker="^",
        color="blue",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCentralDiff,
        label='Central Difference',
        marker="s",
        color="purple",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCtse,
        label='CTSE',
        marker="o",
        color="red",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorDual,
        label='Dual',
        marker="o",
        color="green",
        markerfacecolor="none")
    axes = plt.gca()
    axes.set_xlim([0, 30])
    axes.set_ylim([10**-17, 10**1])
    plt.xlabel('Step size h')
    plt.ylabel('Relative Error')
    xticksOriginal = [0, 5, 10, 15, 20, 25, 30]
    xticksNew = [
        '$10^{0}$',
        '$10^{-5}$',
        '$10^{-10}$',
        '$10^{-15}$',
        '$10^{-20}$',
        '$10^{-25}$',
        '$10^{-30}$']
    plt.xticks(xticksOriginal, xticksNew)
    plt.grid(True)
    plt.legend()
    plt.tight_layout()
    plt.yscale("log")
    plt.show()
    return

def funPlotStepSize_NI(y,h, a_int, b_int, c_param, sInterestedParam, exactDa, exactDb, exactDc, machinePrecision):
    
    import numpy as np
    from matplotlib import pyplot as plt
    import sympy as sym 
    
    #Define the approximation to the integral 
    def I_abc(f,a,b,c):
      y = ((b-a)/8)*(f(a,c) + 3*f((2*a + b)/3,c) + 3*f((a + 2*b)/3,c)  + f(b,c))
      return y
    
    
    
    #Implements CTSE method
    def CTSE(f, a,b,c,h):
     a_pert = a + h*1j
     b_pert = b + h*1j
     c_pert = c + h*1j
     y1 = I_abc(f, a_pert, b, c)
     y2 = I_abc(f, a, b_pert, c)
     y3 = I_abc(f, a,b,c_pert)
     y = y1.real
     dy_da = y1.imag/h
     dy_db = y2.imag/h
     dy_dc = y3.imag/h
     return dy_da, dy_db, dy_dc, y
    
    #Implements FD method
    def fd(f, a,b,c,h):
        dy_da = (I_abc(f, a + h,b,c) - I_abc(f, a,b,c))/h
        dy_db = (I_abc(f, a,b + h,c) - I_abc(f, a,b,c))/h
        dy_dc = (I_abc(f, a,b,c + h) - I_abc(f, a,b,c))/h
        return dy_da, dy_db, dy_dc
    
    #Implements CD method
    def cd(f, a,b,c,h):
        dy_da = (I_abc(f, a + h,b,c) - I_abc(f, a - h,b,c))/(2*h)
        dy_db = (I_abc(f, a,b + h,c) - I_abc(f, a,b - h,c))/(2*h)
        dy_dc = (I_abc(f, a,b,c + h) - I_abc(f, a,b,c - h))/(2*h)
        return dy_da, dy_db, dy_dc
    

    forwardDifference = []
    centralDifference = []
    ctse = []
    
    # Compute f'(x) using forward and central difference, and CTSE.
    for i in h:
        nStepSize = 10**-i
        forwardDifference.append(fd(y, a_int,b_int,c_param,nStepSize))
        centralDifference.append(cd(y, a_int,b_int,c_param,nStepSize ))
        ctse.append(CTSE(y, a_int,b_int,c_param,nStepSize))
        
        
    if sInterestedParam == 'a':
        nChoice = 0
        nExact = exactDa
    if sInterestedParam == 'b':
        nChoice = 1
        nExact = exactDb
    if sInterestedParam == 'c':
        nChoice = 2
        nExact = exactDc
    ctse = np.array(ctse)
    forwardDifference = np.array(forwardDifference)
    centralDifference = np.array(centralDifference)
    ctse = ctse[:,nChoice]
    forwardDifference= forwardDifference[:,nChoice]
    centralDifference= centralDifference[:,nChoice]
 
    # Compute relative errors of the above methods.
    relativeErrorForwardDiff = abs((forwardDifference - nExact) / nExact)
    relativeErrorCentralDiff = abs((centralDifference- nExact) / nExact)
    relativeErrorCtse = abs((ctse  - nExact) / nExact)
    
    mMinVals = [min(relativeErrorForwardDiff ), min(relativeErrorCentralDiff), min(relativeErrorCtse)]
    mMaxVals = [max(relativeErrorForwardDiff ), max(relativeErrorCentralDiff), max(relativeErrorCtse)]
    nMinyVals = min(mMinVals)*10**(-1)
    nMaxyVals = max(mMaxVals)
    # Check if a value in list "relativeErrorCtse" is more precise than
    # machine precision, if so, truncate value to machine precision.
    for i, x in enumerate(relativeErrorCtse):
        if x < machinePrecision:
            relativeErrorCtse[i] = machinePrecision
 
    # Plot and configure figures.
    plt.figure(figsize=(7, 4))
    plt.plot(
        h,
        relativeErrorForwardDiff,
        label='Forward Difference',
        marker="^",
        color="blue",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCentralDiff,
        label='Central Difference',
        marker="s",
        color="purple",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCtse,
        label='CTSE',
        marker="o",
        color="red",
        markerfacecolor="none")
    axes = plt.gca()
    axes.set_xlim([0, 30])
    axes.set_ylim([nMinyVals, nMaxyVals])
    plt.xlabel('Step size h')
    plt.ylabel('Relative Error')
    xticksOriginal = [0, 5, 10, 15, 20, 25, 30]
    xticksNew = [
        '$10^{0}$',
        '$10^{-5}$',
        '$10^{-10}$',
        '$10^{-15}$',
        '$10^{-20}$',
        '$10^{-25}$',
        '$10^{-30}$']
    plt.xticks(xticksOriginal, xticksNew)
    plt.grid(True)
    plt.legend()
    plt.tight_layout()
    plt.yscale("log")
    plt.show()
    
def funPlotStepSize_NI_GL(y,h, a_int, b_int, c_param, sInterestedParam, exactDa, exactDb, exactDc, machinePrecision, w, xi):
    
    import numpy as np
    from matplotlib import pyplot as plt
    import sympy as sym 
    
    #Define the approximation to the integral 
    def I_abc(f,a,b,c, w, xi):
      temp = 0
      for i in range(0,len(w)):
        temp = temp + w[i]*f(((b-a)/2)*xi[i] + ((b+a)/2), c)
      y = ((b - a)/2)*(temp)
      return y
    
    
    
     #Implements CTSE method
    def CTSE(f, a,b,c,h,w, xi):
     a_pert = a + h*1j
     b_pert = b + h*1j
     c_pert = c + h*1j
     y1 = I_abc(f, a_pert, b, c,w,xi)
     y2 = I_abc(f, a, b_pert, c,w,xi)
     y3 = I_abc(f, a,b,c_pert, w, xi)
     y = y1.real
     dy_da = y1.imag/h
     dy_db = y2.imag/h
     dy_dc = y3.imag/h
     return dy_da, dy_db, dy_dc, y
    
    def fd(f, a,b,c,h,w, xi):
        dy_da = (I_abc(f, a + h,b,c, w, xi) - I_abc(f, a,b,c, w, xi))/h
        dy_db = (I_abc(f, a,b + h,c, w, xi) - I_abc(f, a,b,c, w, xi))/h
        dy_dc = (I_abc(f, a,b,c + h, w, xi) - I_abc(f, a,b,c, w, xi))/h
        return dy_da, dy_db, dy_dc
    
    def cd(f, a,b,c,h, w, xi):
        dy_da = (I_abc(f, a + h,b,c,w ,xi) - I_abc(f, a - h,b,c, w, xi))/(2*h)
        dy_db = (I_abc(f, a,b + h,c, w, xi) - I_abc(f, a,b - h,c, w, xi))/(2*h)
        dy_dc = (I_abc(f, a,b,c + h,w, xi) - I_abc(f, a,b,c - h, w, xi))/(2*h)
        return dy_da, dy_db, dy_dc
    

    forwardDifference = []
    centralDifference = []
    ctse = []
    
    # Compute f'(x) using forward and central difference, and CTSE.
    for i in h:
        nStepSize = 10**-i
        forwardDifference.append(fd(y, a_int,b_int,c_param,nStepSize, w, xi))
        centralDifference.append(cd(y, a_int,b_int,c_param,nStepSize, w, xi ))
        ctse.append(CTSE(y, a_int,b_int,c_param,nStepSize, w, xi))
        
        
    if sInterestedParam == 'a':
        nChoice = 0
        nExact = exactDa
    if sInterestedParam == 'b':
        nChoice = 1
        nExact = exactDb
    if sInterestedParam == 'c':
        nChoice = 2
        nExact = exactDc
    ctse = np.array(ctse)
    forwardDifference = np.array(forwardDifference)
    centralDifference = np.array(centralDifference)
    ctse = ctse[:,nChoice]
    forwardDifference= forwardDifference[:,nChoice]
    centralDifference= centralDifference[:,nChoice]
 
    # Compute relative errors of the above methods.
    relativeErrorForwardDiff = abs((forwardDifference - nExact) / nExact)
    relativeErrorCentralDiff = abs((centralDifference- nExact) / nExact)
    relativeErrorCtse = abs((ctse  - nExact) / nExact)
    
    mMinVals = [min(relativeErrorForwardDiff ), min(relativeErrorCentralDiff), min(relativeErrorCtse)]
    mMaxVals = [max(relativeErrorForwardDiff ), max(relativeErrorCentralDiff), max(relativeErrorCtse)]
    nMinyVals = min(mMinVals)*10**(-1)
    nMaxyVals = max(mMaxVals)
    # Check if a value in list "relativeErrorCtse" is more precise than
    # machine precision, if so, truncate value to machine precision.
    for i, x in enumerate(relativeErrorCtse):
        if x < machinePrecision:
            relativeErrorCtse[i] = machinePrecision
 
    # Plot and configure figures.
    plt.figure(figsize=(7, 4))
    plt.plot(
        h,
        relativeErrorForwardDiff,
        label='Forward Difference',
        marker="^",
        color="blue",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCentralDiff,
        label='Central Difference',
        marker="s",
        color="purple",
        markerfacecolor="none")
    plt.plot(
        h,
        relativeErrorCtse,
        label='CTSE',
        marker="o",
        color="red",
        markerfacecolor="none")
    axes = plt.gca()
    axes.set_xlim([0, 30])
    axes.set_ylim([nMinyVals, nMaxyVals])
    plt.xlabel('Step size h')
    plt.ylabel('Relative Error')
    xticksOriginal = [0, 5, 10, 15, 20, 25, 30]
    xticksNew = [
        '$10^{0}$',
        '$10^{-5}$',
        '$10^{-10}$',
        '$10^{-15}$',
        '$10^{-20}$',
        '$10^{-25}$',
        '$10^{-30}$']
    plt.xticks(xticksOriginal, xticksNew)
    plt.grid(True)
    plt.legend()
    plt.tight_layout()
    plt.yscale("log")
    plt.show()    
def funCalcSymbolicFunctions_NI(y,x,a,b,c, a_int, b_int, c_param):
    import sympy as sym
    I = sym.integrate(y,(x,a,b))
    dI_da = sym.diff(I,a)
    dI_db = sym.diff(I,b)
    dI_dc = sym.diff(I,c)
    dI_da = sym.lambdify([a,c], dI_da, 'numpy')
    dI_db = sym.lambdify([b,c], dI_db, 'numpy')
    dI_dc = sym.lambdify([a,b,c], dI_dc, 'numpy')
    exactDa = dI_da(a_int,c_param)
    exactDb = dI_db(b_int,c_param)
    exactDc = dI_dc(a_int,b_int,c_param)
    I = sym.lambdify([a,b,c], I, 'numpy')
    exactFx = I(a_int,b_int,c_param)
    
    
    return  exactDa, exactDb, exactDc, exactFx

# Comment: Should we use given NR subroutines from numpy/scipy 
def NR_method(x0, nIter, nTol, h, f):
  import numpy as np
  from matplotlib import pyplot as plt
  import sympy as sym 
  mVals = []
  mError = []
  for i in range(0,nIter):
    if i == 0:
      x_i = x0 - h*((f(x0 + h*1j)).real)/(f(x0 + h*1j).imag)
      mVals.append(x_i)
      mError.append(np.abs(f(x_i)))
    else:
      x_i =  x_i - h*(((f(x_i + h*1j)).real)/(f(x_i + h*1j).imag))
      mVals.append(x_i)
      mError.append(np.abs(f(x_i)))
      if np.abs(f(x_i)) < nTol:
        break
  plt.figure(figsize=(7, 4))
  plt.plot(mError, marker="o", label = '$|f(x_j)|$',
        color="red",
        markerfacecolor="none")
  axes = plt.gca()
  if i < 10:
    ticks = [j for j in range(0,len(mError))]
    axes.set_xticks(ticks)
  plt.xlabel('Number of iterations')
  plt.ylabel('Forward Error')
  plt.yscale("log")
  plt.grid(True)
  plt.tight_layout()
  plt.hlines(nTol, 0, i, colors='b', linestyles='solid', label='Tolerence')
  plt.legend()
  return



# -------------------------------------------------------------------------------------------------------------------
def NewtonRaphson_FD(f, x0, max_iter=50, tol=1e-6, h=1e-5, full_output=False, f_argv=() ):
    """
    Find a root of a real function using the Newton-Raphson (NR) with Finite Differences.

    INPUTS:
    - f:           callable, Residual function.
    - x0:          float, Initial guess.
    - max_iter:    integer, optional, maximum number of iterations to be performed.
    - tol:         float, optional, tolerance for the residual to terminate the NR loop.
    - step_size:   float, optional, step-size to be used for the CTSE calculations.
    - full_output: bool, optional, request full output.
    - f_argv:       tuple, optional, additional input parameters to evaluate the residual function.

    RESULT:
    The result of this function is a tuple of values that include:
    - x: real, root of the algorithm.
    - n_iter: integer, returned when full_output==True, number of iterations evaluated.
    - x_iter: ndarray, returned when full_output==True, evolution of the root during the iterations.
    - f_iter: ndarray, returned when full_output==True, evolution of the residual function for each iteration.

    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def residual(x):
    >     return np.cos(x) -x
    > # end
    > 
    > x0 = 1.5
    >
    > NewtonRaphson_FD(residual,x0)
    > 0.739085133215161

    """
    import numpy as np
    
    
    x_iter = np.zeros(max_iter) # Reserve memory for max_iter iterations.
    f_iter = np.zeros(max_iter) # Reserve memory for max_iter iterations.
    
    x = x0 # Adjust initial guess.
    n_iter = 0

    for i in range(max_iter):
        
        # Find new evaluation point to perform forward finite differences
        xp = x + h

        # Evaluate residual function
        f_eval    = f(x,*f_argv)
        dfdx_eval = (f(xp,*f_argv)-f_eval)/h

        x = x - f_eval/dfdx_eval # Evaluate Newton-Raphson update.

        # Store information for each iteration.
        x_iter[i] = x
        f_iter[i] = f_eval

        n_iter += 1

        if abs(f_eval.real)<=tol:

            break

        # end if

    # end

    # Check that maximum number of iterations was reached.
    if i == max_iter-1: 
        print('WARNING: Maximum number of iterations reached ({0}).'.format(max_iter))
        print('Check results carefully.')
    else:
        # Resize output arrays to the solved size
        x_iter = x_iter[:n_iter]
        f_iter = f_iter[:n_iter]
    # end if 

    if full_output:
        return ( x, n_iter, x_iter, f_iter )
    else:
        return x
    # end if 

# -------------------------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------------------------
def NewtonRaphson(f, x0, J, max_iter=50, tol=1e-6,  full_output=False, f_argv=() ):
    """
    Find a root of a real function using the Newton-Raphson (NR) with Finite Differences.

    INPUTS:
    - f:           callable, Residual function.
    - x0:          float, Initial guess.
    - J:           callable,  Jacobian of the residual function (J=df/dx). 
    - max_iter:    integer, optional, maximum number of iterations to be performed.
    - tol:         float, optional, tolerance for the residual to terminate the NR loop.
    - full_output: bool, optional, request full output.
    - f_argv:       tuple, optional, additional input parameters to evaluate the residual and jacobian functions.

    RESULT:
    The result of this function is a tuple of values that include:
    - x: real, root of the algorithm.
    - n_iter: integer, returned when full_output==True, number of iterations evaluated.
    - x_iter: ndarray, returned when full_output==True, evolution of the root during the iterations.
    - f_iter: ndarray, returned when full_output==True, evolution of the residual function for each iteration.

    EXAMPLE:

    > import numpy as np
    > from utils.functions import NewtonRaphson
    > 
    > def residual(x):
    >     return np.cos(x) -x
    > # end
    >
    > def jacobian(x):
    >     return -np.sin(x) -1
    > # end
    > 
    > x0 = 1.5
    >
    > NewtonRaphson(residual,x0,jacobian)
    > 0.739085133215161

    """
    import numpy as np
    
    
    x_iter = np.zeros(max_iter) # Reserve memory for max_iter iterations.
    f_iter = np.zeros(max_iter) # Reserve memory for max_iter iterations.
    
    x = x0 # Adjust initial guess.
    n_iter = 0

    for i in range(max_iter):
        
        # Find new evaluation point to perform forward finite differences
        

        # Evaluate residual function
        f_eval    = f(x,*f_argv)

        # Evaluate Jacobian function
        dfdx_eval = J(x,*f_argv)

        x = x - f_eval/dfdx_eval # Evaluate Newton-Raphson update.

        # Store information for each iteration.
        x_iter[i] = x
        f_iter[i] = f_eval

        n_iter += 1

        if abs(f_eval.real)<=tol:

            break

        # end if

    # end

    # Check that maximum number of iterations was reached.
    if i == max_iter-1: 
        print('WARNING: Maximum number of iterations reached ({0}).'.format(max_iter))
        print('Check results carefully.')
    else:
        # Resize output arrays to the solved size
        x_iter = x_iter[:n_iter]
        f_iter = f_iter[:n_iter]
    # end if 

    if full_output:
        return ( x, n_iter, x_iter, f_iter )
    else:
        return x
    # end if 

# -------------------------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------------------------
def NewtonRaphson_Complex(zf, x0, max_iter=50, tol=1e-6, h=1e-10, full_output=False, zf_argv=() ):
    """
    Find a zeros of a real function using the Newton-Raphson (NR) with CTSE.

    INPUTS:
    - zf:          callable, Residual function.
    - x0:          float, Initial guess.
    - max_iter:    integer, optional, maximum number of iterations to be performed.
    - tol:         float, optional, tolerance for the residual to terminate the NR loop.
    - h:           float, optional, step-size to be used for the CTSE calculations.
    - full_output: bool, optional, request full output.

    RESULT:
    The result of this function is a tuple of values that include:
    - x: real, root of the algorithm.
    - n_iter: integer, returned when full_output==True, number of iterations evaluated.
    - x_iter: ndarray, returned when full_output==True, evolution of the root during the iterations.
    - f_iter: ndarray, returned when full_output==True, evolution of the residual function for each iteration.

    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def residual(x):
    >     return np.cos(x) -x
    > # end
    > 
    > x0 = 1.5
    >
    > NewtonRaphson_Complex(residual,x0)
    > 0.739085133215161

    """
    import numpy as np
    
    # Reserve memory for max_iter iterations.
    x_iter = np.zeros(max_iter)
    f_iter = np.zeros(max_iter)
    
    x = x0 # Initial guess.
    n_iter = 0

    # Perform the Newton-Raphson loop.
    for i in range(max_iter):
        
        # Apply CTSE complex perturbation.
        zx = x + h*1j

        # Evaluate residual function
        zf_eval = zf(zx,*zf_argv)

        x = x - h * (zf_eval.real)/(zf_eval.imag) # Evaluate Newton-Raphson update.

        # Store information for each iteration.
        x_iter[i] = x
        f_iter[i] = zf_eval.real

        n_iter += 1

        if abs(zf_eval.real)<=tol:

            break

        # end if

    # end

    # Check that maximum number of iterations was reached.
    if i == max_iter-1: 
        print('WARNING: Maximum number of iterations reached ({0}).'.format(max_iter))
        print('Check results carefully.')
    else:
        # Resize output arrays to the solved size
        x_iter = x_iter[:n_iter]
        f_iter = f_iter[:n_iter]
    # end if 

    if full_output:
        return ( x, n_iter, x_iter, f_iter )
    else:
        return x
    # end if 

# end function
# -------------------------------------------------------------------------------------------------------------------


# -------------------------------------------------------------------------------------------------------------------  
def NewtonConvergence_Plot( x_iter, f_iter, tol, yscale='symlog', xscale='linear', 
        xlabel='Iteration',ylabel='Relative error',grid=True ):
    """
    Generates a StepSize convergence plot with the given data. The data has to be 

    INPUTS:
    - x_iter: ndarray, evolution of the root during the iterations.
    - f_iter: ndarray, evolution of the residual function for each iteration.
    - tol: float, tolerance used for the Newton-Raphson algorithm.
    - yscale: String (optional, default 'symlog'), y-axis scale.
    - xscale: String (optional, default 'symlog'), x-axis scale.
    - xlabel: String (optional, default 'h'), x-axis label.
    - ylabel: String (optional, default 'Relative Error'), y-axis label.
    - grid: Bool (optional, default True), plot grid boolean
    

    RESULT:
    The result of this is a plot showing the behavior of each iteration of the NR algorithm.
    
    """
    import matplotlib.pyplot as plt
    import matplotlib.ticker as ticker
    import numpy as np

    niter = np.arange(1,len(x_iter)+1)
 
    convError = abs((x_iter[1:]-x_iter[:-1])/x_iter[:-1])
    plt.plot(niter[:-1],convError,label='Convergence error of NR solution.')

    # relativeError = abs((x_iter-exactSol)/exactSol)
    # plt.plot(niter,relativeError,label='Relative error of NR solution.')
    
    if yscale == 'symlog':
        plt.yscale(yscale,linthresh=1e-16)
    else:
        plt.yscale(yscale)
    # end if

    # Plot the tolerance as a continous line.
    plt.plot(niter,tol*(np.zeros_like(niter)+1),'--C7',label='Tolerance')
    
    plt.xscale(xscale)

    ax = plt.gca()
    ax.xaxis.set_major_locator(ticker.MultipleLocator(1))

   
    

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    
    plt.legend()


    plt.grid(grid)
    
    return
# -------------------------------------------------------------------------------------------------------------------  

# -------------------------------------------------------------------------------------------------------------------
def CTSE(funct,x,h):
    """
    Evaluates the derivative of a function using the Complex Taylor Series Expansion method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - dfdx: float or array, derivative of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) -x
    > # end
    > 
    > x0 = 1.5
    > h = 1e-20
    >
    > utils.functions.CTSE(function,x0,h)
    > 

    """
    dfdx = (funct(x + h*1j).imag)/h
    return dfdx
# -------------------------------------------------------------------------------------------------------------------  


# -------------------------------------------------------------------------------------------------------------------
def BackwardDiff(funct,x,h):
    """
    Evaluates the derivative of a function using the bakward difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - dfdx: float or array, derivative of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) -x
    > # end
    > 
    > x0 = 1.5
    > h = 1e-4
    >
    > utils.functions.BackwardDiff(function,x0,h)
    > 

    """
    dfdx = ( funct( x ) - funct( x - h ) )/h
    return dfdx
# -------------------------------------------------------------------------------------------------------------------  

# -------------------------------------------------------------------------------------------------------------------
def ForwardDiff_2nd(funct,x,h):
    """
    Evaluates the second derivative of a function using the forward difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - dfdx: float or array, derivative of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) -x
    > # end
    > 
    > x0 = 1.5
    > h = 1e-4
    >
    > utils.functions.CentralDiff(function,x0,h)
    > 

    """
    d2fdx2 = (funct(x + 2*h) -2*funct(x + h) + funct(x))/h**2
    return d2fdx2
# -------------------------------------------------------------------------------------------------------------------  


# -------------------------------------------------------------------------------------------------------------------
def CentralDiff_2nd(funct,x,h):
    """
    Evaluates the second derivative of a function using the central difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - dfdx: float or array, derivative of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) -x
    > # end
    > 
    > x0 = 1.5
    > h = 1e-4
    >
    > utils.functions.ForwardDifference(function,x0,h)
    > 

    """
    d2fdx2 = ( funct(x + h) - 2* funct(x) + funct( x - h ) )/(h*h)
    return d2fdx2
    
# -------------------------------------------------------------------------------------------------------------------  



# -------------------------------------------------------------------------------------------------------------------
def ForwardDiff(funct,x,h):
    """
    Evaluates the derivative of a function using the forward difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - dfdx: float or array, derivative of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) -x
    > # end
    > 
    > x0 = 1.5
    > h = 1e-4
    >
    > utils.functions.CentralDiff(function,x0,h)
    > 

    """
    dfdx = (funct(x + h) - funct(x))/h
    return dfdx
# -------------------------------------------------------------------------------------------------------------------  


# -------------------------------------------------------------------------------------------------------------------
def CentralDiff(funct,x,h):
    """
    Evaluates the derivative of a function using the central difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - dfdx: float or array, derivative of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) -x
    > # end
    > 
    > x0 = 1.5
    > h = 1e-4
    >
    > utils.functions.ForwardDifference(function,x0,h)
    > 

    """
    dfdx = ( funct(x + h) - funct( x - h ) )/(2*h)
    return dfdx
# -------------------------------------------------------------------------------------------------------------------  


def SecondOrderCentralDiff(funct,x,h):
    """
    Evaluates the second derivative of a function using the central difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - d2fdx2: float or array, second derivative of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) -x
    > # end
    > 
    > x0 = 1.5
    > h = 1e-4
    >
    > utils.functions.SecondOrderCentralDiff(function,x0,h)
    > 

    """
    dfdx = ( funct(x + h) - 2*funct(x) + funct( x - h ) )/(h**2)
    return dfdx
# -------------------------------------------------------------------------------------------------------------------  

def SecondOrderCentralDiffBV(funct,x,h):
    """
    Evaluates the second derivatives d^2f/dx^2, d^2f/dy^2 of a two variable function using the central difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - d2fdx2, d2fdy2: float or array, second derivatives of the function at (x,y).
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x,y):
    >     return np.cos(x) - y
    > # end
    > 
    > x0 = [1, 1]
    > h = 1e-4
    >
    > utils.functions.SecondOrderCentralDiffBV(function,x0,h)
    > 

    """
    d2fdx2 = ( funct(x[0] + h, x[1]) - 2*funct(x[0], x[1]) + funct( x[0] - h, x[1] ) )/(h**2)
    d2fdy2 = ( funct(x[0], x[1] + h) - 2*funct(x[0], x[1]) + funct( x[0], x[1] - h ) )/(h**2)
    return d2fdx2, d2fdy2
# -------------------------------------------------------------------------------------------------------------------  

def SecondOrderForwardDiff(funct,x,h):
    """
    Evaluates the second derivative of a single variable function using the forward difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - d2fdx2: float or array, second derivative of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) -x
    > # end
    > 
    > x0 = 1.5
    > h = 1e-4
    >
    > utils.functions.SecondOrderForwardDiff(function,x0,h)
    > 

    """
    dfdx = ( funct(x + 2*h) - 2*funct( x + h ) + funct(x)   )/(h**2)
    return dfdx
# -------------------------------------------------------------------------------------------------------------------  

def SecondOrderForwardDiffBV(funct,x,h):
    """
    Evaluates the second derivatives d^2f/dx^2, d^2f/dy^2 of a two variable function using the forward difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - d2fdx2, d2fdy2: float or array, second derivatives of the function at (x,y).
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) -y
    > # end
    > 
    > x0 = [1,1]
    > h = 1e-4
    >
    > utils.functions.SecondOrderForwardDiffBV(function,x0,h)
    > 

    """
    d2fdx2 = ( funct(x[0] + 2*h, x[1]) - 2*funct( x[0] + h, x[1] ) + funct(x[0], x[1])   )/(h**2)
    d2fdy2 = ( funct(x[0], x[1] + 2*h) - 2*funct(x[0], x[1] + h ) + funct(x[0], x[1])   )/(h**2)
    return d2fdx2, d2fdy2
# -------------------------------------------------------------------------------------------------------------------  


def SecondOrderMixedCentralDiff(funct,x,h):
    """
    Evaluates the second derivatives d^2f/dxdy of a two variable function using a central difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - d2f_dxdy: float or array, mixed second derivative of the function at (x,y).
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) -y
    > # end
    > 
    > x0 = [1,1]
    > h = 1e-4
    >
    > utils.functions.SecondOrderMixedCentralDiffBV(function,x0,h)
    > 

    """
    d2f_dxdy = ( -funct(x[0] + h, x[1] - h) - funct( x[0] - h, x[1] + h) + funct(x[0] + h, x[1]) + funct(x[0] - h, x[1]) + funct(x[0], x[1] + h) + funct(x[0], x[1] - h) - 2*funct(x[0], x[0]))/(2*h**2)
    return d2f_dxdy
# ------------------------------------------------------------------------------------------------------------------- 



def SecondOrderBackwardDiff(funct,x,h):
    """
    Evaluates the second derivative of a function using the backward difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - d2fdx2: float or array, second derivative of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) -x
    > # end
    > 
    > x0 = 1.5
    > h = 1e-4
    >
    > utils.functions.SecondOrderForwardDiff(function,x0,h)
    > 

    """
    dfdx = ( funct(x) - 2*funct( x - h ) + funct(x - 2*h)   )/(h**2)
    return dfdx
# -------------------------------------------------------------------------------------------------------------------  


def SecondOrderBackwardDiffBV(funct,x,h):
    """
    Evaluates the second derivatives d^2f/dx^2, d^2f/dy^2 of a two variable function using the backward difference method.

   INPUTS:
   - funct:   callable, real function.
   - x:       float, scalar value where the derivative is calculated.
   - h:       float or array, step-size(s) to be used for the calculation.

   RESULT:
   The result of this is the same as the value of h (float or array) containing the approxmation of the 
   derivative of the function:
   - d2fdx2, d2fdy2: float or array, second derivatives of the function at (x,y).
   
   EXAMPLE:

   > import numpy as np
   > import utils
   > 
   > def function(x):
   >     return np.cos(x) -y
   > # end
   > 
   > x0 = [1,1]
   > h = 1e-4
   >
   > utils.functions.SecondOrderBackwardDiffBV(function,x0,h)
   > 

    """
    d2fdx2 = ( funct(x[0], x[1]) - 2*funct( x[0] - h, x[1] ) + funct(x[0] - 2*h, x[1])   )/(h**2)
    d2fdy2 = ( funct(x[0], x[1]) - 2*funct( x[0], x[1] - h ) + funct(x[0], x[1] - 2*h)   )/(h**2)
    return d2fdx2, d2fdy2 
# -------------------------------------------------------------------------------------------------------------------  



def CTSEwithForwardDiff(funct,x,h):
    """
    Evaluates the second derivative of a univariate function using CTSE with the forward difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - d2fdx2: float or array, second derivative of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) -x
    > # end
    > 
    > x0 = 1.5
    > h = 1e-4
    >
    > utils.functions.CTSEwithForwardDiff(function,x0,h)
    > 

    """
    
    x1 = x + (10**-10)*1j
    x_real_pert = x + (10**-10)*1j + h
    y = funct(x1)
    y_real_pert = funct(x_real_pert)
    d2f_dx2 =  (y_real_pert.imag - y.imag)/((10**-10)*h)
    return d2f_dx2 
 
# -------------------------------------------------------------------------------------------------------------------  


def CTSEwithForwardDiffBV(funct,x,h):
    """
    Evaluates the second derivatives d^2f/dx^2, d^2f.dy^2, and d^2f/dydx of a bivariate function using CTSE with the forward difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - d2f_dx2, d2f_dy2, d2f_dydx: float or array, second derivatives of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) - y
    > # end
    > 
    > x0 = [1,1]
    > h = 1e-4
    >
    > utils.functions.CTSEwithForwardDiffBV(function,x0,h)
    > 

    """
    
    x1 = x[0] + (10**-10)*1j
    x_real_pert = x[0] + (10**-10)*1j + h
    y = funct(x1, x[1])
    y_real_pert = funct(x_real_pert, x[1])
    d2f_dx2 =  (y_real_pert.imag - y.imag)/((10**-10)*h)
    
    
    y1 = x[1] + (10**-10)*1j
    y_real_pert = x[1] + (10**-10)*1j + h
    y = funct(x[0], y1)
    y_real_pert = funct(x[0], y_real_pert)
    d2f_dy2 =  (y_real_pert.imag - y.imag)/((10**-10)*h)
    
    x1 = x[0] + (10**-10)*1j
    y_real_pert = x[1]  + h
    y = funct(x1, x[1])
    y_real_pert = funct(x1, y_real_pert)
    d2f_dydx =  (y_real_pert.imag - y.imag)/((10**-10)*h)
    
    
    return d2f_dx2, d2f_dy2, d2f_dydx
    

# -------------------------------------------------------------------------------------------------------------------  


def CTSEwithBackwardDiffBV(funct,x,h):
    """
    Evaluates the second derivatives d^2f/dx^2, d^2f/dy^2, d^2f/dydx  of a bivariate function using CTSE with the backward difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - d2f_dx2, d2f_dy2, d2f_dydx: float or array, second derivatives of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) - y
    > # end
    > 
    > x0 = [1,1]
    > h = 1e-4
    >
    > utils.functions.CTSEwithBackwardDiffBV(function,x0,h)
    > 

    """
    
    x1 = x[0] + (10**-10)*1j
    x_real_pert = x[0] + (10**-10)*1j - h
    y = funct(x1, x[1])
    y_real_pert = funct(x_real_pert, x[1])
    d2f_dx2 =  (y.imag - y_real_pert.imag)/((10**-10)*h)
    
    
    y1 = x[1] + (10**-10)*1j
    y_real_pert = x[1] + (10**-10)*1j - h
    y = funct(x[0], y1)
    y_real_pert = funct(x[0], y_real_pert)
    d2f_dy2 =  (y.imag - y_real_pert.imag)/((10**-10)*h)
    
    
    x1 = x[0] + (10**-10)*1j
    y_real_pert = x[1]  - h
    y = funct(x1, x[1])
    y_real_pert = funct(x1, y_real_pert)
    d2f_dydx =   (y.imag - y_real_pert.imag)/((10**-10)*h)
    
    return d2f_dx2, d2f_dy2, d2f_dydx
 
# -------------------------------------------------------------------------------------------------------------------  


def CTSEwithCentralDiff(funct,x,h):
    """
    Evaluates the second derivative of a univariate function using CTSE the central difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - d2fdx2: float or array, second derivative of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) -x
    > # end
    > 
    > x0 = 1.5
    > h = 1e-4
    >
    > utils.functions.SecondOrderForwardDiff(function,x0,h)
    > 

    """
    
    x1 = x + (10**-10)*1j
    x_real_pert_for = x + (10**-10)*1j + h
    x_real_pert_back = x + (10**-10)*1j - h
    y_real_pert_for = funct(x_real_pert_for)
    y_real_pert_back = funct(x_real_pert_back)
    d2f_dx2 =  (y_real_pert_for.imag - y_real_pert_back.imag)/((10**-10)*2*h)
    return d2f_dx2 
 
# -------------------------------------------------------------------------------------------------------------------  


def CTSEwithCentralDiffBV(funct,x,h):
    """
    Evaluates the second derivatives d^2f/dx^2, d^fdy^2, d^2f/dydx of a bivariate function using CTSE with the central difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - d2f_dx2, d2f_dy2, d2f_dydx: float or array, second derivatives of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) - y
    > # end
    > 
    > x0 = [1,1]
    > h = 1e-4
    >
    > utils.functions.CTSEwithCentralDiffBV(function,x0,h)
    > 

    """
    

    x_real_pert_for = x[0] + (10**-10)*1j + h
    x_real_pert_back = x[0] + (10**-10)*1j - h
    y_real_pert_for = funct(x_real_pert_for, x[1])
    y_real_pert_back = funct(x_real_pert_back, x[1])
    d2f_dx2 =  (y_real_pert_for.imag - y_real_pert_back.imag)/((10**-10)*2*h)
    
    
    y_real_pert_for = x[1] + (10**-10)*1j + h
    y_real_pert_back = x[1] + (10**-10)*1j - h
    y_real_pert_for = funct(x[0], y_real_pert_for)
    y_real_pert_back = funct(x[0], y_real_pert_back)
    d2f_dy2 =  (y_real_pert_for.imag - y_real_pert_back.imag)/((10**-10)*2*h)
    
    
    
    x1 = x[0] + (10**-10)*1j
    y_real_pert_back = x[1]  - h
    y_real_pert_for =  x[1]  + h
    y_for = funct(x1,  y_real_pert_for)
    y_back  = funct(x1,  y_real_pert_back)
    d2f_dydx =   (y_for.imag - y_back.imag)/((10**-10)*2*h)
    
    return d2f_dx2, d2f_dy2, d2f_dydx
 
# -------------------------------------------------------------------------------------------------------------------  

def CTSEwithBackwardDiff(funct,x,h):
    """
    Evaluates the second derivative of a univariate function using CTSE with the forward difference method.

    INPUTS:
    - funct:   callable, real function.
    - x:       float, scalar value where the derivative is calculated.
    - h:       float or array, step-size(s) to be used for the calculation.

    RESULT:
    The result of this is the same as the value of h (float or array) containing the approxmation of the 
    derivative of the function:
    - d2fdx2: float or array, second derivative of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) -x
    > # end
    > 
    > x0 = 1.5
    > h = 1e-4
    >
    > utils.functions.SecondOrderForwardDiff(function,x0,h)
    > 

    """
    
    x1 = x + (10**-10)*1j
    x_real_pert = x + (10**-10)*1j - h
    y = funct(x1)
    y_real_pert = funct(x_real_pert)
    d2f_dx2 =  ( y.imag - y_real_pert.imag)/((10**-10)*h)
    return d2f_dx2 


# -------------------------------------------------------------------------------------------------------------------  
def StepSize_Plot( h, derivList, exactDeriv, reversed_xaxis=True, labels=[], yscale='symlog',xscale='log', 
        xlabel='h',ylabel='Relative error',grid=True, markers=[] ):
    """
    Generates a StepSize convergence plot with the given data. The data has to be 

    INPUTS:
    - h: Array, real values of the step sizes used in the study.
    - derivList: List of arrays, each item in this list corresponds to one method.
    - exactDeriv: Float, Exact derivative of the function.
    - reversed_xaxis: Bool (optional, default True), flip x-axis direction.
    - labels: List of strings (optional), lables associated with each item in deriList.
    - yscale: String (optional, default 'symlog'), y-axis scale.
    - xscale: String (optional, default 'symlog'), x-axis scale.
    - xlabel: String (optional, default 'h'), x-axis label.
    - ylabel: String (optional, default 'Relative Error'), y-axis label.
    - grid: Bool (optional, default True), plot grid boolean
    

    RESULT:
    The result of this is a plot showing the behavior of each approximation.
    
    """
    import matplotlib.pyplot as plt

    legend = True
    
    if len(labels) == len(derivList):
        label_list = labels
    elif len(labels) == 0:
        label_list = ['']*len(derivList)
        legend = False
    else:
        raise ValueError( 'The lenght of the labels ({0}) entry should match the number of elements in derivList ({1})'.format(len(labels),len(derivList)))
    # end if

    if len(markers) == len(derivList):
        marker_list = markers
    elif len(markers) == 0:
        marker_list = ['']*len(derivList)
    else:
        raise ValueError( 'The lenght of the markers ({0}) entry should match the number of elements in derivList ({1})'.format(len(markers),len(derivList)))
    # end if

    for i in range(len(derivList)):

        relativeError = abs((derivList[i]-exactDeriv)/exactDeriv)
        plt.plot(h,relativeError,label=label_list[i], marker=marker_list[i])

    # end for
    
    if legend:
        plt.legend()
    # end if 
    
    if yscale == 'symlog':
        plt.yscale(yscale,linthresh=1e-16)
    else:
        plt.yscale(yscale)
    # end if

    plt.xscale(xscale)

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    plt.grid(grid)

    if reversed_xaxis:
        plt.gca().invert_xaxis()
    # end if 
    
    return
# -------------------------------------------------------------------------------------------------------------------  


# -------------------------------------------------------------------------------------------------------------------  
def SymbolicDerivatives(f, x_vals):
    
    # Import symbolic library.
    import sympy as sym
    
    # Creates symbols.
    xs = sym.symbols('x')

    #Evaluates the function symbolically.
    fs = f(xs)

    # Compute the derivative (1st order), and convert it to callable fucntion (lambdify)
    dfdxs = sym.lambdify( xs , fs.diff(xs) )

    # Evaluate the function at the given values of x.
    dfdx_eval = dfdxs(x_vals)

    return dfdx_eval
# -------------------------------------------------------------------------------------------------------------------  



# -------------------------------------------------------------------------------------------------------------------  
def SingleVar_ClosedForm( f, labels=[],  xlabel='x',ylabel='f(x)', grid=True ):
    """
    Generates a StepSize convergence plot with the given data. The data has to be 

    INPUTS:
    - f: Callable, function to be evaluated.
    

    RESULT:
    The result of this is a plot showing the behavior of each approximation.
    - dfdx: float or array, derivative of the function at x.
    
    EXAMPLE:

    > import numpy as np
    > import utils
    > 
    > def function(x):
    >     return np.cos(x) -x
    > # end
    > 
    > x0 = 1.5
    > h = 1e-4
    >
    > utils.functions.ForwardDifference(function,x0,h)
    > 
    """
    import matplotlib.pyplot as plt

    legend = True
    
    if len(labels) == len(derivList):
        label_list = labels
    elif len(labels) == 0:
        label_list = ['']*len(derivList)
        legend = False
    else:
        raise ValueError( 'The lenght of the labels ({0}) entry should match the number of elements in derivList ({1})'.format(len(labels),len(derivList)))
    # end if

    for i in range(len(derivList)):

        relativeError = abs((derivList[i]-exactDeriv)/exactDeriv)
        plt.plot(h,relativeError,label=label_list[i])

    # end for
    
    if legend:
        plt.legend()
    # end if 
    
    if yscale == 'symlog':
        plt.yscale(yscale,linthresh=1e-16)
    else:
        plt.yscale(yscale)
    # end if

    plt.xscale(xscale)

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    plt.grid(grid)

    if reversed_xaxis:
        plt.gca().invert_xaxis()
    # end if 
    
    return
# -------------------------------------------------------------------------------------------------------------------  
